import unittest

from watergrid.context import DataContext
from watergrid.pipelines.pipeline import Pipeline
from watergrid.steps import Step


class MockStep(Step):
    def __init__(self):
        super().__init__(self.__class__.__name__)
        self.mock_flag = False

    def run(self, context: DataContext):
        self.mock_flag = True

    def get_flag(self):
        return self.mock_flag


class PipelineTestCase(unittest.TestCase):
    def test_create_pipeline(self):
        pipeline = Pipeline("test_pipeline")
        self.assertIsNotNone(pipeline)

    def test_pipeline_runs_step(self):
        step = MockStep()
        pipeline = Pipeline("test_pipeline")
        pipeline.add_step(step)
        pipeline.run()
        self.assertTrue(step.get_flag())

    def test_pipeline_runs_multiple_steps(self):
        step1 = MockStep()
        step2 = MockStep()
        pipeline = Pipeline("test_pipeline")
        pipeline.add_step(step1)
        pipeline.add_step(step2)
        pipeline.run()
        self.assertTrue(step1.get_flag())
        self.assertTrue(step2.get_flag())


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod

from watergrid.context import DataContext


class Step(ABC):
    """
    Abstract class for a single step in a pipeline.

    :param step_name: Name of the step.
    :type step_name: str
    :param provides: List of data keys that this step provides.
    :type provides: list
    :param requires: List of data keys that this step requires.
    :type requires: list
    """

    def __init__(self, step_name: str, provides: list = None, requires: list = None):
        self._step_name = step_name

        if provides is None:
            self.__dep_provides = []
        else:
            self.__dep_provides = provides

        if requires is None:
            self.__dep_requires = []
        else:
            self.__dep_requires = requires

    def run_step(self, context: DataContext):
        """
        Used internally by the pipeline. Performs setup and teardown
        in addition to running the step function run().
        :return: None
        """
        self.run(context)

    def get_step_requirements(self) -> list:
        """
        Returns a list of data keys that this step requires.
        :return: List of data keys that this step requires.
        :rtype: list
        """
        return self.__dep_requires

    def get_step_provides(self) -> list:
        """
        Returns a list of data keys that this step provides.
        :return: List of data keys that this step provides.
        :rtype: list
        """
        return self.__dep_provides

    def get_step_name(self) -> str:
        """
        Returns the name of the step.
        :return: Name of the step.
        """
        return self._step_name

    @abstractmethod
    def run(self, context: DataContext):
        pass

    def __str__(self) -> str:
        return self._step_name



import unittest

from watergrid.context import DataContext
from watergrid.metrics.ConsoleMetricsExporter import ConsoleMetricsExporter
from watergrid.pipelines.pipeline import Pipeline
from watergrid.steps import Step


class MockStep(Step):
    def __init__(self, throw_exception=False):
        self.throw_exception = throw_exception
        super().__init__(self.__class__.__name__)

    def run(self, context: DataContext):
        if self.throw_exception:
            raise Exception("MockStep failed")


class ConsoleMetricsExporterTestCase(unittest.TestCase):
    def test_outputs_pipeline_start(self):
        pipeline = Pipeline("test_pipeline")
        pipeline.add_metrics_exporter(ConsoleMetricsExporter())
        with self.assertLogs() as captured:
            pipeline.run()
        self.assertIn("INFO:root:Starting pipeline: test_pipeline", captured.output)

    def test_outputs_pipeline_end(self):
        pipeline = Pipeline("test_pipeline")
        pipeline.add_metrics_exporter(ConsoleMetricsExporter())
        with self.assertLogs() as captured:
            pipeline.run()
        self.assertIn("INFO:root:Ending pipeline", captured.output)

    def test_outputs_pipeline_end_with_error(self):
        pipeline = Pipeline("test_pipeline")
        pipeline.add_metrics_exporter(ConsoleMetricsExporter())
        pipeline.add_step(MockStep(throw_exception=True))
        with self.assertLogs() as captured:
            pipeline.run()
        self.assertIn("ERROR:root:Exception: MockStep failed", captured.output)

    def test_outputs_step_start(self):
        pipeline = Pipeline("test_pipeline")
        pipeline.add_metrics_exporter(ConsoleMetricsExporter())
        pipeline.add_step(MockStep())
        with self.assertLogs() as captured:
            pipeline.run()
        self.assertIn("INFO:root:Starting step: MockStep", captured.output)

    def test_outputs_step_end(self):
        pipeline = Pipeline("test_pipeline")
        pipeline.add_metrics_exporter(ConsoleMetricsExporter())
        pipeline.add_step(MockStep())
        with self.assertLogs() as captured:
            pipeline.run()
        self.assertIn("INFO:root:Ending step", captured.output)


if __name__ == "__main__":
    unittest.main()



from abc import ABC

from watergrid.steps import Step


class Sequence(ABC):
    """
    A sequence object is used to form a logical grouping of steps. It can be
    used to simplify your pipeline initialization code, or it can be used to
    rapidly re-use multiple steps at once.
    """

    def __init__(self, name: str):
        self.name = name
        self.steps = []

    def add_step(self, step: Step) -> None:
        """
        Adds a step to the sequence.
        :param step: Step to add.
        :return: None
        """
        self.steps.append(step)

    def export_steps(self) -> list:
        """
        Returns a list of all steps in the sequence. Used internally by the
        pipeline class.
        :return: List of all steps.
        """
        return self.steps



import unittest

from watergrid.context import DataContext
from watergrid.metrics.MetricsExporter import MetricsExporter
from watergrid.pipelines.pipeline import Pipeline
from watergrid.steps import Step


class MockMetricsExporter(MetricsExporter):
    def start_pipeline(self, pipeline_name):
        self.__pipeline_start = True

    def end_pipeline(self):
        self.__pipeline_end = True

    def start_step(self, step_name):
        self.__step_start += 1

    def end_step(self):
        self.__step_end += 1

    def get_pipeline_start(self):
        return self.__pipeline_start

    def get_pipeline_end(self):
        return self.__pipeline_end

    def get_step_start(self):
        return self.__step_start

    def get_step_end(self):
        return self.__step_end

    def get_excpetion_raised(self):
        return self.__exception_raised

    def capture_exception(self, exception: Exception):
        self.__exception_raised = True

    def __init__(self, step_count):
        self.__pipeline_start = False
        self.__pipeline_end = False
        self.__step_start = 0
        self.__step_end = 0
        self.__step_count = step_count
        self.__exception_raised = False
        super().__init__()


class MockStep(Step):
    def __init__(self, throw_exception=False):
        super().__init__(self.__class__.__name__)
        self.mock_flag = False
        self.throw_exception = throw_exception

    def run(self, context: DataContext):
        self.mock_flag = True
        if self.throw_exception:
            raise Exception("Mock exception")

    def get_flag(self):
        return self.mock_flag


class PipelineMetricsTestCase(unittest.TestCase):
    def test_pipeline_accepts_metrics(self):
        pipeline = Pipeline("test_pipeline")
        exporter = MockMetricsExporter(1)
        pipeline.add_metrics_exporter(exporter)
        step1 = MockStep()
        pipeline.add_step(step1)
        pipeline.run()
        self.assertTrue(step1.get_flag())
        self.assertTrue(exporter.get_pipeline_start())
        self.assertTrue(exporter.get_pipeline_end())
        self.assertEqual(1, exporter.get_step_start())
        self.assertEqual(1, exporter.get_step_end())
        self.assertFalse(exporter.get_excpetion_raised())

    def test_pipeline_measures_two_steps(self):
        pipeline = Pipeline("test_pipeline")
        exporter = MockMetricsExporter(2)
        pipeline.add_metrics_exporter(exporter)
        step1 = MockStep()
        step2 = MockStep()
        pipeline.add_step(step1)
        pipeline.add_step(step2)
        pipeline.run()
        self.assertTrue(step1.get_flag())
        self.assertTrue(step2.get_flag())
        self.assertTrue(exporter.get_pipeline_start())
        self.assertTrue(exporter.get_pipeline_end())
        self.assertEqual(2, exporter.get_step_start())
        self.assertEqual(2, exporter.get_step_end())
        self.assertFalse(exporter.get_excpetion_raised())

    def test_metrics_captures_exception(self):
        pipeline = Pipeline("test_pipeline")
        exporter = MockMetricsExporter(1)
        pipeline.add_metrics_exporter(exporter)
        step1 = MockStep(True)
        pipeline.add_step(step1)
        pipeline.run()
        self.assertTrue(step1.get_flag())
        self.assertTrue(exporter.get_pipeline_start())
        self.assertTrue(exporter.get_pipeline_end())
        self.assertEqual(1, exporter.get_step_start())
        self.assertEqual(1, exporter.get_step_end())
        self.assertTrue(exporter.get_excpetion_raised())


if __name__ == "__main__":
    unittest.main()



import unittest

from watergrid.context import DataContext
from watergrid.pipelines.pipeline import Pipeline
from watergrid.steps import Step


class MockStepA(Step):
    def __init__(self):
        super().__init__(self.__class__.__name__)
        self.mock_flag = False
        self.last_run_flag = False

    def run(self, context: DataContext):
        if context.get_run_metadata() is not None:
            self.mock_flag = True
            if context.get_run_metadata().get_last_pipeline_run() is not None:
                self.last_run_flag = True

    def get_context_flag(self):
        return self.mock_flag

    def get_last_run_flag(self):
        return self.last_run_flag


class ContextMetaTestCase(unittest.TestCase):
    def test_meta_initializes(self):
        context = DataContext()
        self.assertIsNotNone(context.get_run_metadata())

    def test_can_use_meta_in_step(self):
        pipeline = Pipeline("test_pipeline")
        step1 = MockStepA()
        pipeline.add_step(step1)
        pipeline.run()
        self.assertTrue(step1.get_context_flag())

    def test_can_use_meta_in_step_with_last_run(self):
        pipeline = Pipeline("test_pipeline")
        step1 = MockStepA()
        pipeline.add_step(step1)
        pipeline.run()
        pipeline.run()
        self.assertTrue(step1.get_last_run_flag())


if __name__ == "__main__":
    unittest.main()



import unittest

from watergrid.context import DataContext, OutputMode
from watergrid.pipelines.pipeline import Pipeline
from watergrid.steps import Step


class TestDTO:
    def __init__(self, value):
        self.value = value

    def set_value(self, value):
        self.value = value

    def get_value(self):
        return self.value


class TestCreateDTOStep(Step):
    def run(self, context: DataContext):
        context.set("dto", TestDTO(1))
        context.set("val", [1, 2])
        context.set_output_mode(OutputMode.SPLIT)

    def __init__(self):
        super().__init__("test_create_dto_step", provides=["val", "dto"])


class TestModifyDTOStep(Step):
    def __init__(self):
        super().__init__("test_modify_dto_step", requires=["val"])

    def run(self, context: DataContext):
        if context.get("val") == 1:
            context.get("dto").set_value(5)


class TestVerifyCopySafetyStep(Step):
    def __init__(self):
        super().__init__("test_verify_copy_safety_step", requires=["val"])
        self.mod_1_flag = -1
        self.keep_2_flag = -1

    def run(self, context: DataContext):
        if context.get("val") == 1:
            self.mod_1_flag = context.get("dto").get_value()
        elif context.get("val") == 2:
            self.keep_2_flag = context.get("dto").get_value()


class Bug6TestCase(unittest.TestCase):
    def test_bug6(self):
        pipeline = Pipeline("test_pipeline")
        pipeline.add_step(TestCreateDTOStep())
        pipeline.add_step(TestModifyDTOStep())
        step3 = TestVerifyCopySafetyStep()
        pipeline.add_step(step3)
        pipeline.run()
        self.assertEqual(5, step3.mod_1_flag)
        self.assertEqual(1, step3.keep_2_flag)



import base64
import copy
import time
from abc import ABC

import pycron

from watergrid.context import DataContext, OutputMode, ContextMetadata
from watergrid.metrics.MetricsStore import MetricsStore
from watergrid.pipelines.pipeline_verifier import PipelineVerifier
from watergrid.steps import Sequence
from watergrid.steps import Step


class Pipeline(ABC):
    """
    Internal class that represents a pipeline of steps.

    :param pipeline_name: Name of the pipeline.
    :type pipeline_name: str
    """

    def __init__(self, pipeline_name: str):
        self._pipeline_name = pipeline_name
        self._steps = []
        self._metrics_store = MetricsStore()
        self._last_run = None

    def add_step(self, step: Step):
        """
        Add a step to the pipeline.

        :param step: Step to add.
        :type step: Step
        """
        self._steps.append(step)

    def add_steps(self, steps: Sequence):
        """
        Adds a sequence of steps to the pipeline.
        :param steps: Sequence implementation that contains the steps to add.
        :return: None
        """
        for step in steps.export_steps():
            self.add_step(step)

    def run(self):
        """
        Blocking operation that runs all steps in the pipeline once.
        :return: None
        """
        self.verify_pipeline()
        self._metrics_store.start_pipeline_monitoring(self._pipeline_name)
        try:
            self.__run_pipeline_steps()
            self._last_run = time.time()
        except Exception as e:
            self._metrics_store.report_exception(e)
            self._metrics_store.stop_step_monitoring()
        finally:
            self._metrics_store.stop_pipeline_monitoring()

    def run_loop(self):
        """
        Runs the pipeline in a loop. Subsequent executions will run immediately after the previous execution.
        """
        while True:
            self.run()

    def run_scheduled(self, cron_expression: str):
        """
        Runs the pipeline once every time the cron expression matches.

        :param cron_expression: Cron expression to match.
        :type cron_expression: str
        """
        while True:
            if pycron.is_now(cron_expression):
                start_time = time.perf_counter()
                self.run()
                elapsed_time = time.perf_counter() - start_time
                time.sleep(
                    60 - (elapsed_time / 1000)
                )  # Sleep for the remaining time in the minute so that the pipeline does not run multiple times per cron trigger.
            else:
                time.sleep(1)

    def get_step_count(self):
        """
        :return: Number of steps in the pipeline.
        """
        return len(self._steps)

    def get_pipeline_name(self) -> str:
        """
        :return: Name of the pipeline.
        """
        return self._pipeline_name

    def verify_pipeline(self):
        """
        Verifies that the pipeline is valid and that all step dependencies are met.
        :return: None
        """
        PipelineVerifier.verify_pipeline_dependencies_fulfilled(self._steps)
        PipelineVerifier.verify_pipeline_step_ordering(self._steps)

    def add_metrics_exporter(self, exporter):
        """
        Adds a metrics exporter to the pipeline metric store.
        :param exporter: Exporter to add.
        :return: None
        """
        self._metrics_store.add_metrics_exporter(exporter)

    def get_pipeline_guid(self):
        """
        Generates a unique identifier for the pipeline that can be used to
        identify all pipelines that have the same steps in the same order with
        the same name.
        :return: GUID of the pipeline.
        """
        result = self._pipeline_name
        for step in self._steps:
            result += step.get_step_name() + type(step).__name__
        return base64.urlsafe_b64encode(result.encode("utf-8"))

    def __run_pipeline_steps(self):
        """
        Performs setup and runs all steps in the pipeline.
        :return:
        """
        contexts = [self.__generate_first_context()]
        for step in self._steps:
            contexts = self.__run_step_for_each_context(step, contexts)
            if len(contexts) == 0:
                break

    def __generate_first_context(self) -> DataContext:
        """
        Generates the first context for the pipeline.
        :return: First context.
        """
        context = DataContext()
        context_meta = ContextMetadata(self._last_run)
        context.set_run_metadata(context_meta)
        return context

    def __run_step_for_each_context(self, step: Step, context_list: list) -> list:
        """
        Runs the given step once for each context in the context list and performs post-processing.
        :param step: Step to run.
        :param context_list: List of contexts to provide to the step runtime.
        :return: List of contexts to provide to the next step.
        """
        next_contexts = []
        for context in context_list:
            self.__run_step_with_performance_monitoring(step, context)
            self.__process_step_output(context, step.get_step_provides(), next_contexts)
        return next_contexts

    def __process_step_output(
        self, context: DataContext, step_provides: list, next_contexts: list
    ):
        """
        Performs post-processing on the output of a step.
        :param context: Output from the current step.
        :param step_provides: List of data keys that the step provides.
        :param next_contexts: List of contexts that will be passed to the next step.
        :return: None
        """
        if context.get_output_mode() == OutputMode.SPLIT:
            self.__split_context(step_provides, context, next_contexts)
        elif context.get_output_mode() == OutputMode.FILTER:
            self.__filter_context(step_provides, context, next_contexts)
        else:
            self.__forward_context(context, next_contexts)

    def __run_step_with_performance_monitoring(self, step: Step, context: DataContext):
        """
        Runs a single step in the pipeline and forwards performance data to any installed monitoring plugins.
        :param step: Step to run.
        :param context: Context to provide to the step runtime.
        :return: None
        """
        self._metrics_store.start_step_monitoring(step.get_step_name())
        step.run_step(context)
        self._metrics_store.stop_step_monitoring()

    def __split_context(
        self, step_provides: list, context: DataContext, next_contexts: list
    ):
        """
        Splits the context into multiple contexts based on the output of the current step.
        :param step_provides: The list of data keys that the current step provides.
        :param context: The context instance from the current step.
        :param next_contexts: The list of contexts that will be passed to the next step.
        :return: None
        """
        split_key = step_provides[0]
        split_value = context.get(split_key)
        for value in split_value:
            new_context = copy.deepcopy(context)
            new_context.set_output_mode(OutputMode.DIRECT)
            new_context.set(split_key, value)
            next_contexts.append(new_context)

    def __filter_context(
        self, step_provides: list, context: DataContext, next_contexts: list
    ):
        """
        Forwards the context to the next step if the given filter key is present.
        :param step_provides: The list of context keys that the step provides.
        :param context: The context instance from the current step.
        :param next_contexts: The list of contexts that will be passed to the next step.
        :return: None
        """
        if context.get(step_provides[0]) is not None:
            new_context = copy.deepcopy(context)
            new_context.set_output_mode(OutputMode.DIRECT)
            next_contexts.append(new_context)

    def __forward_context(self, context: DataContext, next_contexts: list):
        """
        Forwards the context to the next step with no post-processing.
        :param context: Context to forward.
        :param next_contexts: List of contexts that will be used by the next step.
        :return: None
        """
        next_contexts.append(copy.deepcopy(context))



from abc import ABC, abstractmethod


class MetricsExporter(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def start_pipeline(self, pipeline_name):
        pass

    @abstractmethod
    def end_pipeline(self):
        pass

    @abstractmethod
    def start_step(self, step_name):
        pass

    @abstractmethod
    def end_step(self):
        pass

    @abstractmethod
    def capture_exception(self, exception: Exception):
        pass



import time
from abc import ABC, abstractmethod


class PipelineLock(ABC):
    def __init__(self, lock_timeout: int = 60):
        self.lock_timeout = lock_timeout
        self.__start_time = None
        self.__has_lock = False

    def lock(self) -> bool:
        if self.has_lock():
            raise RuntimeError("Lock already acquired")
        if self.acquire():
            self.__start_time = time.perf_counter()
            self.__has_lock = True
            return True
        return False

    def ping(self):
        if not self.has_lock():
            raise RuntimeError("Lock not acquired")
        if not self.has_lock():
            raise RuntimeError("Lock was lost")
        lock_time = time.perf_counter() - self.__start_time
        if lock_time > (self.lock_timeout * 0.9):
            self.extend_lease()

    def unlock(self):
        if not self.has_lock():
            raise RuntimeError("Lock not acquired")
        self.release()
        self.__has_lock = False

    @abstractmethod
    def acquire(self) -> bool:
        pass

    @abstractmethod
    def has_lock(self) -> bool:
        pass

    @abstractmethod
    def extend_lease(self):
        pass

    @abstractmethod
    def release(self):
        pass

    @abstractmethod
    def read_key(self, key: str):
        pass

    @abstractmethod
    def write_key(self, key: str, value):
        pass



import logging

from watergrid.metrics.MetricsExporter import MetricsExporter


class ConsoleMetricsExporter(MetricsExporter):
    def __init__(self):
        super().__init__()
        logging.basicConfig(level=logging.INFO)
        self._logger = logging.getLogger(__name__)

    def start_pipeline(self, pipeline_name):
        logging.info("Starting pipeline: " + pipeline_name)

    def end_pipeline(self):
        logging.info("Ending pipeline")

    def start_step(self, step_name):
        logging.info("Starting step: " + step_name)

    def end_step(self):
        logging.info("Ending step")

    def capture_exception(self, exception: Exception):
        logging.error("Exception: " + str(exception))



import unittest

from watergrid.pipelines.standalone_pipeline import StandalonePipeline


class StandalonePipelineTestCases(unittest.TestCase):
    def test_can_initialize_pipeline(self):
        pipeline = StandalonePipeline("test_pipeline")
        self.assertEqual(pipeline.get_pipeline_name(), "test_pipeline")



from redis import Redis
from redis.lock import Lock

from watergrid.locks.PipelineLock import PipelineLock


class RedisPipelineLock(PipelineLock):
    def __init__(self, lock_timeout: int = 60):
        super().__init__(lock_timeout=lock_timeout)
        self._lock_key = "pipeline-lock"
        self._redis_host = "localhost"
        self._redis_port = 6379
        self._redis_db = 0
        self._redis_password = None
        self.__redis = None
        self.__lock = None

    def connect(self):
        self.__redis = Redis(
            host=self._redis_host,
            port=self._redis_port,
            db=self._redis_db,
            password=self._redis_password,
        )
        self.__redis.ping()
        self.__lock = Lock(self.__redis, self._lock_key, timeout=self.lock_timeout)

    def set_host(self, host: str):
        self._redis_host = host

    def set_port(self, port: int):
        self._redis_port = port

    def set_db(self, db: int):
        self._redis_db = db

    def set_password(self, password: str):
        self._redis_password = password

    def acquire(self) -> bool:
        return self.__lock.acquire(blocking=False)

    def has_lock(self) -> bool:
        return self.__lock.owned()

    def extend_lease(self):
        self.__lock.extend(self.lock_timeout)

    def release(self):
        self.__lock.release()

    def read_key(self, key: str):
        return self.__redis.get(key)

    def write_key(self, key: str, value: str):
        self.__redis.set(key, value)



import logging
import time

from watergrid.locks.PipelineLock import PipelineLock
from watergrid.pipelines.pipeline import Pipeline


class HAPipeline(Pipeline):
    """
    The high availability pipeline allows for several machines to have the pipeline loaded at once. If one of the
    machines fails, or if the pipeline context times out, the pipeline will run on another machine.

    :param pipeline_name: The name of the pipeline.
    :param pipeline_lock: The lock implementation to use for this pipeline.
    """

    def __init__(self, pipeline_name: str, pipeline_lock: PipelineLock):
        super().__init__(pipeline_name)
        self.__pipeline_lock = pipeline_lock
        self.__lock_timings = []

    def run(self):
        self.verify_pipeline()
        if self.__pipeline_lock.lock():
            super().run()
            self.__pipeline_lock.unlock()
        else:
            logging.debug(
                "Pipeline {} is already running on another instance".format(
                    self.get_pipeline_name()
                )
            )

    def run_interval(self, job_interval_s: int) -> None:
        """
        Runs the pipeline in a blocking interval mode. Every job_interval_s seconds, the pipeline is run.
        :param job_interval_s: Number of seconds to wait between pipeline runs.
        :return: None
        """
        self.verify_pipeline()
        while True:
            self._run_interval_loop(job_interval_s)
            time.sleep(self._calculate_delay(job_interval_s) / 1000)

    def _run_interval_loop(self, job_interval_s: int) -> None:
        """
        Runs the pipeline in interval mode once. If the lock is successfully acquired, the pipeline steps are run.
        :param job_interval_s: Number of seconds to wait between pipeline runs.
        :return: None
        """
        if self.__pipeline_lock.lock():
            self._perform_locked_interval_actions(job_interval_s)
            self.__pipeline_lock.unlock()
        else:
            self._handle_lock_acquire_failure()

    def _perform_locked_interval_actions(self, job_interval_s: int) -> None:
        """
        Performs the actions that should be performed when the pipeline is locked. Note that this method assumes that
        the lock has already been acquired and is held by the calling class/function.
        :param job_interval_s: Number of seconds to wait between pipeline runs.
        :return: None
        """
        self._verify_lock_metadata(job_interval_s)
        last_run = self._get_last_run()
        if time.time() - last_run > job_interval_s:
            super().run()
            self._set_last_run(last_run + job_interval_s)

    def _verify_lock_metadata(self, job_interval_s: int) -> None:
        """
        Verifies that the lock metadata is correct. If the metadata is not correct, the lock is deleted and recreated.
        :param job_interval_s: Number of seconds to wait between pipeline runs.
        :return: None
        """
        last_run = self._get_last_run()
        if last_run == 0:
            self._set_last_run(time.time() - job_interval_s)
        if time.time() - last_run > job_interval_s * 3:
            logging.warning(
                "Pipeline {} has fallen more than three cycles behind. Consider increasing the job interval or "
                "provisioning more machines.".format(self.get_pipeline_name())
            )
            self._set_last_run(time.time() - job_interval_s)

    def _handle_lock_acquire_failure(self) -> None:
        """
        Handles the case where the pipeline lock could not be acquired. This is a no-op for the HA pipeline. Does not
        run if the lock could not be acquired due to an exception.
        :return: None
        """
        logging.debug(
            "Pipeline {} is already running on another instance".format(
                self.get_pipeline_name()
            )
        )

    def _set_last_run(self, last_run: float) -> None:
        """
        Sets the timestamp of the last run of this pipeline as recorded in the lock backend.
        :param last_run: Timestamp to record as the last run of this pipeline.
        :return: None
        """
        self.__pipeline_lock.write_key(self._get_pipeline_lock_name(), last_run)

    def _get_last_run(self) -> float:
        """
        Gets the timestamp of the last run of this pipeline as recorded in the lock backend. Note that this value
        keeps the interval defined by job_interval_s, and may not represent the exact time of the last run.
        :return: Timestamp of the last run of this pipeline.
        """
        try:
            return float(self.__pipeline_lock.read_key(self._get_pipeline_lock_name()))
        except TypeError:
            return 0

    def _get_pipeline_lock_name(self) -> str:
        """
        Builds the name of the lock assigned to this pipeline.
        :return: Pipeline lock name.
        """
        return "{}_last_run".format(self.get_pipeline_name())

    def lock_with_timing(self) -> None:
        """
        Acquire the pipeline lock and record the time it took to acquire the lock to the lock timings counter.
        :return: None
        """
        start_time = time.perf_counter()
        self.__pipeline_lock.lock()
        self._append_timing(time.perf_counter() - start_time)

    def unlock_with_timing(self) -> None:
        """
        Unlocks the pipeline and appends the time taken to unlock to the lock timings counter.
        :return: None
        """
        start_time = time.perf_counter()
        self.__pipeline_lock.unlock()
        self._append_timing(time.perf_counter() - start_time)

    def get_average_lock_delay(self) -> float:
        """
        Returns the average time it takes to perform a lock operation with the currently configured backend lock.
        :return: Average operation time in milliseconds.
        """
        if len(self.__lock_timings) == 0:
            return 0
        return float(sum(self.__lock_timings)) / float(len(self.__lock_timings))

    def _calculate_delay(
        self,
        pipeline_interval_s: int,
        checks_per_interval: int = 10,
        check_ratio: int = 3,
    ) -> int:
        """
        Calculates the delay in milliseconds to wait before running the pipeline again. This is based on the configured
        pipeline interval, and the average performance of the backend hosting the pipeline lock.
        :param pipeline_interval_s: The interval in seconds between pipeline runs.
        :param checks_per_interval: Number of times the lock should be chedked per interval across all nodes.
        :param check_ratio: Should be a number close to the average number of pipeline nodes.
        :return: Delay in milliseconds to wait before running the pipeline again.
        """
        redis_delay_ms = self.get_average_lock_delay()
        job_delay_ms = float(pipeline_interval_s * 1000) / float(checks_per_interval)
        if redis_delay_ms > job_delay_ms:
            logging.warning(
                "Slow redis cluster detected. Consider increasing the size of your cluster."
            )
        return int(max(redis_delay_ms, job_delay_ms)) * check_ratio

    def _append_timing(self, timing: float) -> None:
        """
        Adds a lock operation benchmark timing to the rolling average counter. Used for calculating mutex lookup delays.
        :param timing: The time in milliseconds it took to perform the lock operation.
        :return: None
        """
        self.__lock_timings.append(timing)
        if len(self.__lock_timings) > 100:
            self.__lock_timings.pop(0)



import unittest

from watergrid.context import DataContext
from watergrid.pipelines.pipeline import Pipeline
from watergrid.steps import Step


class MockStep(Step):
    def __init__(self, provides=[], requires=[]):
        super().__init__(self.__class__.__name__, provides, requires)
        if provides is None:
            provides = []
        self.run_order = 0

    def run(self, context: DataContext):
        if context.has("test_key"):
            self.run_order = context.get("test_key") + 1
            context.set("test_key", self.run_order)
        else:
            context.set("test_key", 1)
            self.run_order = 1

    def get_flag(self):
        return self.run_order


class PipelineSchedulerTestCase(unittest.TestCase):
    def test_pipeline_does_not_break_in_order_steps(self):
        pipeline = Pipeline("test_pipeline")
        step_a = MockStep(provides=["a"])
        pipeline.add_step(step_a)
        step_b = MockStep(requires=["a"])
        pipeline.add_step(step_b)
        pipeline.run()
        self.assertEqual(1, step_a.get_flag())
        self.assertEqual(2, step_b.get_flag())

    def test_pipeline_does_not_delete_steps(self):
        pipeline = Pipeline("test_pipeline")
        step_a = MockStep(provides=["a"])
        pipeline.add_step(step_a)
        step_b = MockStep(requires=["a"])
        pipeline.add_step(step_b)
        pipeline.run()
        self.assertEqual(2, pipeline.get_step_count())

    def test_pipeline_throws_if_unmet_dependencies(self):
        pipeline = Pipeline("test_pipeline")
        step_b = MockStep(requires=["a"])
        pipeline.add_step(step_b)
        with self.assertRaises(Exception):
            pipeline.run()
        self.assertEqual(0, step_b.get_flag())

    def test_pipeline_schedules_basic_dependencies_in_steps(self):
        pipeline = Pipeline("test_pipeline")
        step_a = MockStep(provides=["a"])
        step_b = MockStep(requires=["a"])
        pipeline.add_step(step_b)
        pipeline.add_step(step_a)
        pipeline.run()
        self.assertEqual(pipeline.get_step_count(), 2)
        self.assertEqual(step_a.get_flag(), 1)
        self.assertEqual(step_b.get_flag(), 2)

    def test_pipeline_schedules_three_dependencies(self):
        pipeline = Pipeline("test_pipeline")
        step_a = MockStep(provides=["a"])
        step_b = MockStep(requires=["a"], provides=["b"])
        step_c = MockStep(requires=["b"])
        pipeline.add_step(step_c)
        pipeline.add_step(step_b)
        pipeline.add_step(step_a)
        pipeline.run()
        self.assertEqual(3, pipeline.get_step_count())
        self.assertEqual(1, step_a.get_flag())
        self.assertEqual(2, step_b.get_flag())
        self.assertEqual(3, step_c.get_flag())

    def test_pipeline_throw_error_on_circular_dependencies(self):
        pipeline = Pipeline("test_pipeline")
        step_a = MockStep(provides=["a"], requires=["b"])
        step_b = MockStep(provides=["b"], requires=["a"])
        pipeline.add_step(step_a)
        pipeline.add_step(step_b)
        with self.assertRaises(Exception):
            pipeline.run()


if __name__ == "__main__":
    unittest.main()



from watergrid.metrics.MetricsExporter import MetricsExporter


class MetricsStore:
    def __init__(self):
        self._metrics_exporters = []

    def add_metrics_exporter(self, metrics_exporter: MetricsExporter):
        self._metrics_exporters.append(metrics_exporter)

    def start_pipeline_monitoring(self, pipeline_name: str):
        for exporter in self._metrics_exporters:
            exporter.start_pipeline(pipeline_name)

    def stop_pipeline_monitoring(self):
        for exporter in self._metrics_exporters:
            exporter.end_pipeline()

    def start_step_monitoring(self, step_name: str):
        for exporter in self._metrics_exporters:
            exporter.start_step(step_name)

    def stop_step_monitoring(self):
        for exporter in self._metrics_exporters:
            exporter.end_step()

    def report_exception(self, exception: Exception):
        for exporter in self._metrics_exporters:
            exporter.capture_exception(exception)



import unittest

from watergrid.context import DataContext
from watergrid.pipelines.pipeline import Pipeline
from watergrid.steps import Step


class MockStep(Step):
    def __init__(self, name):
        super().__init__(name)

    def run(self, context: DataContext):
        pass


class MockStep2(Step):
    def __init__(self, name):
        super().__init__(name)

    def run(self, context: DataContext):
        pass


class PipelineVersioningTestCase(unittest.TestCase):
    def test_pipeline_has_guid_function(self):
        pipeline = Pipeline("test_pipeline")
        self.assertIsNotNone(pipeline.get_pipeline_guid())

    def test_pipeline_guid_is_unique_by_name(self):
        pipeline1 = Pipeline("test_pipeline1")
        pipeline2 = Pipeline("test_pipeline2")
        self.assertNotEqual(
            pipeline1.get_pipeline_guid(), pipeline2.get_pipeline_guid()
        )

    def test_pipeline_guid_is_unique_by_steps(self):
        pipeline1 = Pipeline("test_pipeline")
        pipeline1.add_step(MockStep("step1"))
        pipeline2 = Pipeline("test_pipeline")
        pipeline2.add_step(MockStep("step2"))
        self.assertNotEqual(
            pipeline1.get_pipeline_guid(), pipeline2.get_pipeline_guid()
        )

    def test_pipeline_guid_is_unique_by_step_class_names(self):
        pipeline1 = Pipeline("test_pipeline")
        pipeline1.add_step(MockStep("step1"))
        pipeline2 = Pipeline("test_pipeline")
        pipeline2.add_step(MockStep2("step1"))
        self.assertNotEqual(
            pipeline1.get_pipeline_guid(), pipeline2.get_pipeline_guid()
        )

    def test_pipeline_has_clean_guid(self):
        pipeline = Pipeline("test_pipeline")
        self.assertNotIn(str(pipeline.get_pipeline_guid()), " ")


if __name__ == "__main__":
    unittest.main()



import unittest

from watergrid.context import DataContext
from watergrid.pipelines.pipeline import Pipeline
from watergrid.steps import Sequence
from watergrid.steps import Step


class MockStep(Step):
    def __init__(self):
        super().__init__("mock_step")
        self.flag = False

    def run(self, context: DataContext):
        self.flag = True

    def get_flag(self) -> bool:
        return self.flag


class TestSequence(Sequence):
    def __init__(self):
        super().__init__("test_sequence")


class SequenceTestCase(unittest.TestCase):
    def test_sequence_adds_steps(self):
        pipeline = Pipeline("test_pipeline")
        mock_step = MockStep()
        sequence = TestSequence()
        sequence.add_step(mock_step)
        pipeline.add_steps(sequence)
        pipeline.run()
        self.assertTrue(mock_step.get_flag())


if __name__ == "__main__":
    unittest.main()



import time
import unittest

from watergrid.context import DataContext
from watergrid.locks import MockPipelineLock
from watergrid.pipelines.ha_pipeline import HAPipeline
from watergrid.steps import Step


class MockStep(Step):
    def __init__(self):
        super().__init__(self.__class__.__name__)
        self.mock_flag = False

    def run(self, context: DataContext):
        self.mock_flag = True

    def get_flag(self):
        return self.mock_flag


class HAPipelineTestCase(unittest.TestCase):
    def test_pipeline_does_not_runs_without_lock(self):
        step1 = MockStep()
        step2 = MockStep()
        mock_lock = MockPipelineLock()
        pipeline = HAPipeline("test_pipeline", mock_lock)
        pipeline.add_step(step1)
        pipeline.add_step(step2)
        mock_lock.manual_lock()
        pipeline.run()
        self.assertFalse(step1.get_flag())
        self.assertFalse(step2.get_flag())

    def test_pipeline_runs_with_lock(self):
        step1 = MockStep()
        step2 = MockStep()
        mock_lock = MockPipelineLock()
        pipeline = HAPipeline("test_pipeline", mock_lock)
        pipeline.add_step(step1)
        pipeline.add_step(step2)
        pipeline.run()
        self.assertTrue(step1.get_flag())
        self.assertTrue(step2.get_flag())

    def test_pipeline_releases_lock(self):
        step1 = MockStep()
        step2 = MockStep()
        mock_lock = MockPipelineLock()
        pipeline = HAPipeline("test_pipeline", mock_lock)
        pipeline.add_step(step1)
        pipeline.add_step(step2)
        pipeline.run()
        self.assertFalse(mock_lock.has_lock())

    def test_pipeline_calculates_redis_delay(self):
        pipeline = HAPipeline("test_pipeline", MockPipelineLock())
        pipeline._append_timing(10000)
        pipeline._append_timing(20000)
        self.assertEqual(15000, pipeline.get_average_lock_delay())

    def test_pipeline_calculates_job_interval_delay(self):
        pipeline = HAPipeline("test_pipeline", MockPipelineLock())
        self.assertEqual(3000, pipeline._calculate_delay(10))

    def test_pipline_does_not_run_interval_without_lock(self):
        step1 = MockStep()
        pipeline_lcok = MockPipelineLock()
        pipeline = HAPipeline("test_pipeline", pipeline_lcok)
        pipeline.add_step(step1)
        pipeline_lcok.manual_lock()
        pipeline._run_interval_loop(10)
        self.assertFalse(step1.get_flag())

    def test_pipeline_generates_lock_name(self):
        pipeline = HAPipeline("test_pipeline", MockPipelineLock())
        self.assertEqual("test_pipeline_last_run", pipeline._get_pipeline_lock_name())

    def test_pipeline_waits_for_target_time(self):
        step1 = MockStep()
        pipeline_lock = MockPipelineLock()
        pipeline = HAPipeline("test_pipeline", pipeline_lock)
        pipeline.add_step(step1)
        pipeline._set_last_run(time.time() + 30)
        pipeline._run_interval_loop(10)
        self.assertFalse(step1.get_flag())

    def test_pipeline_runs_interval_with_target_time(self):
        step1 = MockStep()
        pipeline_lock = MockPipelineLock()
        pipeline = HAPipeline("test_pipeline", pipeline_lock)
        pipeline.add_step(step1)
        pipeline._set_last_run(time.time() - 20)
        pipeline._run_interval_loop(10)
        self.assertTrue(step1.get_flag())

    def test_pipeline_sets_metadata(self):
        step1 = MockStep()
        pipeline_lock = MockPipelineLock()
        pipeline = HAPipeline("test_pipeline", pipeline_lock)
        pipeline.add_step(step1)
        pipeline._run_interval_loop(10)
        self.assertIsNotNone(pipeline._get_last_run())

    def test_pipeline_locks_with_benchmarking(self):
        pipeline_lock = MockPipelineLock()
        pipeline = HAPipeline("test_pipeline", pipeline_lock)
        pipeline.lock_with_timing()
        self.assertTrue(pipeline_lock.has_lock())
        pipeline.unlock_with_timing()
        self.assertFalse(pipeline_lock.has_lock())
        self.assertNotEqual(0, pipeline.get_average_lock_delay())


if __name__ == "__main__":
    unittest.main()



from watergrid.locks.PipelineLock import PipelineLock


class MockPipelineLock(PipelineLock):
    def read_key(self, key: str):
        try:
            return self.__key_value[key]
        except KeyError:
            return None

    def write_key(self, key: str, value: str):
        self.__key_value[key] = value

    def __init__(self, lock_timeout: int = 60):
        super().__init__(lock_timeout)
        self.external_lock = False
        self.client_lock = False
        self.__key_value = {}

    def manual_lock(self):
        self.external_lock = True
        self.client_lock = False

    def manual_unlock(self):
        self.external_lock = False

    def acquire(self) -> bool:
        if not self.external_lock:
            self.client_lock = True
            return True
        else:
            return False

    def has_lock(self) -> bool:
        return self.client_lock

    def extend_lease(self):
        pass

    def release(self):
        self.client_lock = False



from watergrid.steps import Step


class PipelineVerifier:
    @staticmethod
    def verify_pipeline_dependencies_fulfilled(pipeline_steps: list):
        """
        Verifies that all dependencies of the pipeline are fulfilled by at least one other step. Does not check
        validity of step ordering in the pipeline.
        :param pipeline_steps: List of all steps in the pipeline.
        :return: None
        """
        # Get a list of all provided data keys.
        provided_keys = PipelineVerifier.__get_all_step_provides(pipeline_steps)
        # Check that all dependencies are met.
        PipelineVerifier.__check_for_unlinked_dependencies(
            provided_keys, pipeline_steps
        )

    @staticmethod
    def __get_all_step_provides(pipeline_steps: list) -> list:
        """
        Gets a list of all unique keys provided by all steps in the pipeline.
        :param pipeline_steps: List of all steps in the pipeline.
        :return: List of all data keys provided by all steps in the pipeline.
        """
        provided_keys = []
        for step in pipeline_steps:
            provided_keys.extend(
                PipelineVerifier.__get_unique_step_provides(step, provided_keys)
            )
        return provided_keys

    @staticmethod
    def __get_unique_step_provides(step: Step, provided_keys: list) -> list:
        """
        Gets a list of all unique keys provided by a single step.
        :param step: Step to check.
        :param provided_keys: List of all data keys provided by all steps in the pipeline.
        :return: List of all data keys provided by the step.
        """
        new_keys = []
        for key in step.get_step_provides():
            if key not in provided_keys:
                new_keys.append(key)
        return new_keys

    @staticmethod
    def __check_for_unlinked_dependencies(
        provided_keys: list, pipeline_steps: list
    ) -> None:
        """
        Checks that all dependencies of the pipeline are fulfilled by at least one other step. Does not check
        validity of step ordering in the pipeline.
        :param provided_keys: List of all data keys provided by all steps in the pipeline.
        :param pipeline_steps: List of all steps in the pipeline.
        :return: None
        """
        for step in pipeline_steps:
            PipelineVerifier.__check_step_dependencies(step, provided_keys)

    @staticmethod
    def __check_step_dependencies(step: Step, provided_keys: list) -> None:
        """
        Checks that all dependencies of a single step are fulfilled by at least one other step.
        :param step: Step to check.
        :param provided_keys: List of all data keys provided by all steps in the pipeline.
        :return: None
        """
        for step_dependency in step.get_step_requirements():
            if step_dependency not in provided_keys:
                raise Exception(
                    f"Step {step.get_step_name()} requires {step_dependency} to be provided."
                )

    @staticmethod
    def verify_pipeline_step_ordering(pipeline_steps: list) -> None:
        """
        Verifies that all steps in the pipeline are in the correct order.
        :return: None
        """
        # Get a list of all provided data keys.
        provided_keys = []
        step_index = 0
        index_hop_count = 0
        while step_index < len(pipeline_steps):
            index_hop_count += 1
            PipelineVerifier.__recursive_loop_failsafe(
                index_hop_count, len(pipeline_steps)
            )
            if PipelineVerifier.__check_step_position_is_correct(
                pipeline_steps, step_index, provided_keys
            ):
                step_index += 1

    @staticmethod
    def __recursive_loop_failsafe(hop_count: int, total_steps: int) -> None:
        if hop_count > total_steps**2:
            raise Exception("Unable to resolve step dependencies.")

    @staticmethod
    def __check_step_position_is_correct(
        pipeline_steps: list, step_index: int, provided_keys: list
    ) -> bool:
        if not PipelineVerifier.__bump_step_if_dependency_not_met(
            pipeline_steps, step_index, provided_keys
        ):
            provided_keys.extend(
                PipelineVerifier.__get_unique_step_provides(
                    pipeline_steps[step_index], provided_keys
                )
            )
            return True
        return False

    @staticmethod
    def __bump_step_if_dependency_not_met(
        pipeline_steps: list, step_index: int, provided_keys: list
    ) -> bool:
        for step_dependency in pipeline_steps[step_index].get_step_requirements():
            if step_dependency not in provided_keys:
                pipeline_steps.append(pipeline_steps.pop(step_index))
                return True
        return False



from watergrid.pipelines.pipeline import Pipeline


class StandalonePipeline(Pipeline):
    """
    A composable pipeline that runs on a single host. Provides no high availability or fault tolerance.

    :param name: The name of the pipeline.
    :type name: str
    """

    def __init__(self, pipeline_name: str):
        super().__init__(pipeline_name)



from watergrid.context import OutputMode
from watergrid.context.context_metadata import ContextMetadata


class DataContext:
    """
    Stores data from previous steps, and allows you to pass KV pairs to
    subsequent steps.
    """

    def __init__(self):
        self.data = {}
        self.output_mode = OutputMode.DIRECT
        self.metadata = ContextMetadata()

    def set(self, key: str, value: object) -> None:
        """
        Sets a key-value pair in the pipeline context to be passed to another step.
        :param key: Key to set.
        :param value: Value to assign to the given key.
        :return: None
        """
        self.data[key] = value

    def get(self, key: str):
        """
        Gets the value of a given key.
        :param key: Key of the value to get.
        :return: Value of the given key (if found in the KV store).
        """
        return self.data[key]

    def get_all(self) -> dict:
        """
        Gets all the data stored in the pipeline context.
        :return: Dictionary of all KV pairs in the pipeline context.
        """
        return self.data

    def set_batch(self, batch: dict) -> None:
        """
        Overwrites all data in the pipeline context with the given dict.
        :param batch: Data to overwrite the pipeline context with.
        :return: None
        """
        self.data = batch

    def has(self, key: str) -> bool:
        """
        Checks if a given key is in the pipeline context.
        :param key: Key to check for.
        :return: Boolean denoting if the key was found.
        """
        return key in self.data

    def set_output_mode(self, mode: OutputMode) -> None:
        """
        Sets the output mode of the pipeline context. Determines if the context will be deleted, forwarded,
        or split before the next step.
        :param mode: Mode to set the pipeline context to.
        :return: None
        """
        self.output_mode = mode

    def reset_context(self):
        """
        Resets the pipeline context mode to the default.
        :return:
        """
        self.output_mode = OutputMode.DIRECT

    def get_output_mode(self) -> OutputMode:
        """
        Gets the output mode of the pipeline context.
        :return: Currently configured output mode of the pipeline context.
        """
        return self.output_mode

    def get_run_metadata(self) -> ContextMetadata:
        """
        Gets the metadata associated with the current run.
        :return: Metadata associated with the current run.
        """
        return self.metadata

    def set_run_metadata(self, metadata: ContextMetadata) -> None:
        """
        Sets the metadata associated with the current run.
        :param metadata: Metadata to associate with the current run.
        :return: None
        """
        self.metadata = metadata

    @staticmethod
    def deep_copy_context(context):
        """
        Creates a deep copy of a DataContext object.
        :param context: Context instance to be copied.
        :return: New copy of the given context instance.
        """
        new_context = DataContext()
        new_context.set_batch(dict(context.get_all()))
        return new_context



import unittest

from watergrid.context import DataContext, OutputMode
from watergrid.pipelines.pipeline import Pipeline
from watergrid.steps import Step


class MockSplitStep(Step):
    def __init__(self, test_data: list):
        super().__init__(self.__class__.__name__, provides=["test_data"])
        self.__test_data = test_data

    def run(self, context: DataContext):
        context.set("test_data", self.__test_data)
        context.set_output_mode(OutputMode.SPLIT)


class MockFilterStep(Step):
    def __init__(self):
        super().__init__(
            self.__class__.__name__,
            provides=["filtered_test_data"],
            requires=["test_data"],
        )

    def run(self, context: DataContext):
        test_data = context.get("test_data")
        context.set_output_mode(OutputMode.FILTER)
        if test_data == 2:
            context.set("filtered_test_data", test_data)
        else:
            context.set("filtered_test_data", None)


class MockStep(Step):
    def __init__(self):
        super().__init__(self.__class__.__name__)
        self.mock_flag = 0

    def run(self, context: DataContext):
        self.mock_flag += 1

    def get_flag(self):
        return self.mock_flag


class MockVerifyStep(Step):
    def __init__(self, test_key: str):
        super().__init__(self.__class__.__name__)
        self.mock_flag = 0
        self.__test_key = test_key

    def run(self, context: DataContext):
        self.mock_flag += context.get(self.__test_key)

    def get_flag(self):
        return self.mock_flag


class DataContextTestCase(unittest.TestCase):
    def test_normal_context_runs_once(self):
        step1 = MockStep()
        step2 = MockStep()
        pipeline = Pipeline("test_pipeline")
        pipeline.add_step(step1)
        pipeline.add_step(step2)
        pipeline.run()
        self.assertEqual(1, step1.get_flag())
        self.assertEqual(1, step2.get_flag())

    def test_split_context_runs_twice(self):
        step1 = MockSplitStep(test_data=[0, 1])
        step2 = MockStep()
        pipeline = Pipeline("test_pipeline")
        pipeline.add_step(step1)
        pipeline.add_step(step2)
        pipeline.run()
        self.assertEqual(2, step2.get_flag())

    def test_split_context_splits_data(self):
        step1 = MockSplitStep(test_data=[1, 3, 5, 7, 9])
        step2 = MockVerifyStep("test_data")
        pipeline = Pipeline("test_pipeline")
        pipeline.add_step(step1)
        pipeline.add_step(step2)
        pipeline.run()
        self.assertEqual(25, step2.get_flag())

    def test_filter_context_removes_values(self):
        step1 = MockSplitStep(test_data=[1, 2])
        step2 = MockFilterStep()
        step3 = MockVerifyStep("filtered_test_data")
        pipeline = Pipeline("test_pipeline")
        pipeline.add_step(step1)
        pipeline.add_step(step2)
        pipeline.add_step(step3)
        pipeline.run()
        self.assertEqual(2, step3.get_flag())


if __name__ == "__main__":
    unittest.main()



import unittest

import watergrid


class PackageTestCase(unittest.TestCase):
    def test_version(self):
        self.assertNotEqual(watergrid.__version__, "")


if __name__ == "__main__":
    unittest.main()



import time
import unittest

from watergrid.locks import RedisPipelineLock


class PipelineTestCase(unittest.TestCase):
    def test_can_connect(self):
        redis_lock = RedisPipelineLock()
        redis_lock.connect()

    def test_can_cycle_lock(self):
        redis_lock = RedisPipelineLock()
        redis_lock.connect()
        redis_lock.lock()
        self.assertTrue(redis_lock.has_lock())
        redis_lock.unlock()

    def test_can_set_host(self):
        redis_lock = RedisPipelineLock()
        redis_lock.set_host("localhost")

    def test_can_set_port(self):
        redis_lock = RedisPipelineLock()
        redis_lock.set_port(6379)

    def test_can_set_db(self):
        redis_lock = RedisPipelineLock()
        redis_lock.set_db(0)

    def test_can_set_password(self):
        redis_lock = RedisPipelineLock()
        redis_lock.set_password("password")

    def test_can_renew_lock(self):
        redis_lock = RedisPipelineLock()
        redis_lock.connect()
        redis_lock.lock()
        redis_lock.extend_lease()
        redis_lock.unlock()

    def test_can_write_keys(self):
        redis_lock = RedisPipelineLock()
        redis_lock.connect()
        redis_lock.write_key("key", "value3")
        self.assertEqual("value3", redis_lock.read_key("key").decode("utf-8"))

    def test_can_write_timestamps(self):
        redis_lock = RedisPipelineLock()
        timestamp = time.time()
        redis_lock.connect()
        redis_lock.write_key("timekey", timestamp)
        self.assertEqual(
            timestamp, float(redis_lock.read_key("timekey").decode("utf-8"))
        )



import setuptools
import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="watergrid",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Joshua Zenn",
    description="Lightweight framework for building ETL pipelines.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ARMmaster17/watergrid-python",
    packages=setuptools.find_packages(),
    package_dir={"watergrid": "watergrid"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pycron==3.0.0",
    ],
    extras_require={
        "lock-redis": ["redis==4.2.2"],
        "metrics-elasticsearch": ["elastic-apm==6.9.1"],
    },
    setup_requires=["black==22.3.0"],
)



class ContextMetadata:
    def __init__(self, last_pipeline_run_timestamp=None):
        self._last_pipeline_run = last_pipeline_run_timestamp

    def get_last_pipeline_run(self) -> float:
        """
        Gets the timestamp of the last pipeline run. The time represents the current time when the last step finished.
        :return: Timestamp of last pipeline run.
        """
        return self._last_pipeline_run



import unittest

from watergrid.locks import MockPipelineLock


class PipelineLockTests(unittest.TestCase):
    def test_can_acquire_lock(self):
        lock = MockPipelineLock()
        lock.lock()
        self.assertTrue(lock.has_lock())

    def test_can_release_lock(self):
        lock = MockPipelineLock()
        lock.lock()
        lock.unlock()
        self.assertFalse(lock.has_lock())

    def test_cannot_release_missing_lock(self):
        lock = MockPipelineLock()
        with self.assertRaises(RuntimeError):
            lock.unlock()

    def test_cannot_double_lock(self):
        lock = MockPipelineLock()
        lock.lock()
        with self.assertRaises(RuntimeError):
            lock.lock()

    def test_can_ping_acquired_lock(self):
        lock = MockPipelineLock()
        lock.lock()
        lock.ping()

    def test_ping_fails_on_lock_lost(self):
        lock = MockPipelineLock()
        lock.lock()
        lock.manual_lock()
        with self.assertRaises(RuntimeError):
            lock.ping()

    def test_cannot_ping_missing_lock(self):
        lock = MockPipelineLock()
        with self.assertRaises(RuntimeError):
            lock.ping()


if __name__ == "__main__":
    unittest.main()



from watergrid.metrics.MetricsExporter import MetricsExporter


class ElasticAPMMetricsExporter(MetricsExporter):
    def __init__(self, app_name: str, apm_server_url: str, apm_secret_token: str):
        super().__init__()
        self.__client = Client(
            service_name=app_name,
            server_url=apm_server_url,
            secret_token=apm_secret_token,
        )
        instrument()
        self.__pipeline_failed = False
        self.__pipeline_name = None

    def start_pipeline(self, pipeline_name: str):
        self.__client.begin_transaction(transaction_type="script")
        self.__pipeline_name = pipeline_name

    def end_pipeline(self):
        if self.__pipeline_failed:
            elasticapm.set_transaction_outcome(OUTCOME.FAILURE)
            self.__client.end_transaction(self.__pipeline_name, "failure")
            self.__pipeline_failed = False
        else:
            elasticapm.set_transaction_outcome(OUTCOME.SUCCESS)
            self.__client.end_transaction(self.__pipeline_name, "success")

    def start_step(self, step_name: str):
        pass  # This is handled by the instrumentation functions of Elastic APM.

    def end_step(self):
        pass  # This is handled by the instrumentation functions of Elastic APM.

    def capture_exception(self, e: Exception):
        self.__client.capture_exception(
            exc_info=(type(e), e, e.__traceback__), handled=True
        )
        self.__pipeline_failed = True
